/*



      JRG, 5/25/03          **** E. O. P. ****
                     *** EMU Offline Processing ***
	   (based on old program CRC_CALC, 8 May 2003 version )
     ROOT-based histogramming added by Valeri Andreev and P. J. Murray

To run:
- Type the executable, eop2, at the command line and press enter.
- The computer will then ask for "Run Name" before processing; it will then      look for the file "run{RunName}.dat" but RunName must be a string less than    80 characters.
- You can also, at the command line, type the executable, space, then the "Run   Name".
- If you don't want all of the output printed to the screen, after the "Run      Name" type a space and then >out.


Functions:
              check CRCs and wordcounts for ALCT, TMB, CFEB
	         -N_CFEB_TimeSamples is taken into account (CFEB bit 13)
              check L1As and BXNs for ALCT, TMB, CFEB
              count empty events, CFEB events, ALCT event, TMB events
              count board errors for multiple CSCs: use ierr_xxx for dmb, tmb
              count number of single- and multi-CSC events (DMB HDRs found)
	      checks run for proper control code and readout protocol
              decodes status/error codes from DDU
To Do:
  for Key CFEB Hit histo, make 6 bins; 1-5 for CFEBs, 0 for "empty"
     -maybe also make a histo for Generic CFEB Hit (not just key)

  histogram everything (!) especially BXN
     -#CFEBs/ALCTs/TMBs/DMBs, fully-occupied CSCs
     -errors per DDU Input channel, per DMB/TMB/ALCT?
     -event size (DDU-WC), #packets?
     -histo for every DDU status/error bit?
  add BINARY DATA WRITE for Excellent events
  add BINARY DATA READ for future DDU-DAQ format
  add detection of 16-bit offset in FIFO stream, also correction?
  add voting for DDU Special bits
*/

#include <stdio.h>
#include <time.h>
#include <fcntl.h>
#include <unistd.h>  /* sleep function */
#include <signal.h>  /* Usual interrupt handler stuff */
#include <string.h>
// V.A. begin
#include <iostream>
#include <string>  
#include "TFile.h"
#include "TH1.h"
#include "TH2.h"
#include "TProfile.h"
#include "TNtuple.h"
#include "TRandom.h"

/* For DDU 32-bit Error/Status Decode */
void ddu_status_decode(int long code)
{
// printf("\nReceived code=%08X\n",code);
// if((code&0x008FFFFF)>0){
/* Non-triggered^^^^^ status bits:
	      bit[31:28] = Private gigabit Ethernet status
	      bit[27:24] = details related to a triggered error
	      bit[23:21] = less important status/warning bits
*/

// JRG, low-order 16-bit status (most serious errors):
 if ((code&0x0000F000) > 0){
   if((0x00008000&code)>0)printf("   DDU Critical Error, ** needs reset **\n");
   if ((0x00004000&code) > 0) printf("   DDU Single Error, bad event");
   if ((0x00002000&code) > 0) printf("   DDU Single Warning");
   if ((0x00001000&code) > 0) printf("   DDU Near Full Warning");
   printf("\n");
 }
 if ((code&0x00000F00) > 0){
   if ((0x00000800&code) > 0)printf("   DDU RX Error");
   if ((0x00000400&code) > 0)printf("   DDU Control DLL Error occured");
   if ((0x00000200&code) > 0)printf("   DDU DMB Error occurred");
   if ((0x00000100&code) > 0)printf("   DDU Lost In Event Error");
   printf("\n");
 }
 if ((code&0x000000F0) > 0){
   if ((0x00000080&code) > 0) printf("   DDU Lost In Data Error occurred");
   if ((0x00000040&code) > 0) printf("   DDU Timeout Error");

// Multiple-bit vote failures (or Rx Errors) in one 64-bit word:

   if ((0x00000020&code) > 0) printf("   DDU Critical Data Error");

// Multiple single-bit vote failures (or Rx Errors) over time from one DMB:
  
   if ((0x00000010&code) > 0) printf("   DDU Multiple Transmit Errors");
   printf("\n");
 }
 if ((code&0x0000000F) > 0){
   if ((0x00000008&code) > 0) printf("   DDU FIFO Full Error");
   if ((0x00000004&code) > 0) printf("   DDU Fiber Error");
   if ((0x00000002&code) > 0) printf("   DDU L1A Match Error");
   if ((0x00000001&code) > 0) printf("   DDU CRC Error");
   printf("\n");
 }
 if ((code&0xF0000000) > 0){

// JRG, high-order 16-bit status (not-so-serious errors):

/*Pre-ddu2ctrl_ver51 encoding:
   if ((0x80000000&code) > 0) printf("   DDU G-Bit FIFO Not Empty");
   if ((0x40000000&code) > 0) printf("   DDU G-Bit FIFO Near Full Warning");
   if ((0x20000000&code) > 0) printf("   DDU G-Bit FIFO Full Warning");
*/

   if ((0x80000000&code) > 0) printf("   DDU Output Limited Buffer Overflow");
   if ((0x40000000&code) > 0) printf("   DDU G-Bit FIFO Full Warning");
   if ((0x20000000&code) > 0) printf("   DDU G-Bit FIFO Near Full Warning");
   if ((0x10000000&code) > 0) printf("   DDU G-Bit Fiber Error");
   printf("\n");
 }
 if ((code&0x0F000000) > 0){
   if ((0x08000000&code) > 0)printf("   DDU FirstDat Error");

// Pre-ddu3ctrl_v8r15576:   if((0x04000000&code)>0)printf("   DDU BadFirstWord Error");

   if ((0x04000000&code) > 0) printf("   DDU L1A-FIFO Full Error");

// Pre-ddu2ctrl_ver53:   if((0x02000000&code)>0)printf("   DDU BadCtrlWord Error");

   if ((0x02000000&code) > 0) printf("   DDU Data Stuck in FIFO");
   if ((0x01000000&code) > 0) printf("   DDU NoLiveFibers Error");
   printf("\n");
 }
 if ((code&0x00F00000) > 0){
   if ((0x00800000&code) > 0) printf("   DDU Spwd single-bit Warning");
   if ((0x00400000&code) > 0) printf("   DDU Ethernet DLL Error");
   if ((0x00200000&code) > 0) printf("   DDU S-Link Full Bit set");
   if ((0x00100000&code) > 0) printf("   DDU S-Link Not Ready");
   if ((0x00300000&code) == 0x00200000) printf("\n     DDU S-Link Stopped (backpressure)");
   printf("\n");
 }
 if ((code&0x000F0000) > 0){
   if ((0x00080000&code) > 0) printf("   DDU TMB Error");

/*Pre-ddu2ctrl_ver57:
   if ((0x00040000&code) > 0) printf("   DDU FIFO-PAF Warning");
   if ((0x00020000&code) > 0) printf("   DDU L1A-FIFO Near Full Warning");
*/

/*
Pre-ddu3ctrl_v8r15576:
   if ((0x00040000&code) > 0) printf("   DDU Trigger Readout CRC Error");
   if ((0x00020000&code) > 0) printf("   DDU Trigger Readout Wordcount Error");
   if ((0x00010000&code) > 0) printf("   DDU L1A-FIFO Full Error");
*/
   if ((0x00040000&code) > 0) printf("   DDU ALCT Error");
   if ((0x00020000&code) > 0) printf("   DDU Trigger Readout Wordcount Error");
   if ((0x00010000&code) > 0) printf("   DDU Trigger L1A Match Error");
   printf("\n");
 }
}

void ADChistos(unsigned short bigarray[16][96], TH2F *phisto1, TH2F *phisto2, unsigned int numtimesa, unsigned short cflag){

  float max = 0;
  unsigned int incre, timecount;
  float baseline;
  unsigned int aflag;

 for (incre=0;incre<96;incre++){
   aflag = 0;
   //   baseline = (((bigarray[0][incre])&0x0fff) + ((bigarray[1][incre])&0x0fff))/2;   
   baseline = ((bigarray[0][incre])&0x0fff);
   for (timecount=0;timecount<numtimesa;timecount++){
     bigarray[timecount][incre] = (bigarray[timecount][incre]&0x0fff);
     if (bigarray[timecount][incre] > (1.5*baseline)) aflag = 1;
     if (max < bigarray[timecount][incre]) max = bigarray[timecount][incre];
   }
   //   printf("%6.1f = Max, %6.1f = baseline, %1d = flag, %1d = chamber\n", max, baseline, aflag, cflag);
 for (timecount=0;timecount<numtimesa;timecount++){
     if (aflag == 1){
       //  printf("Timesample: %d",timecount);
       // printf("       Array value: %d\n",bigarray[timecount][incre]);
	 bigarray[timecount][incre] = bigarray[timecount][incre] - baseline;
	 if (cflag == 0) phisto1->Fill(timecount,((1.0*bigarray[timecount][incre])/(max-baseline)));
	 if (cflag == 1) phisto2->Fill(timecount,((1.0*bigarray[timecount][incre])/(max-baseline)));
     }
 }
 max = 0;
 }
}


// Main program starts here.

main(int argc, char *argv[]){
  int ierr,i,j,k,l,m,n,t,num,ev=0,buf[4],l1offset=0,prev_offset,bx_offset=0;
  int trg_bx_offset=0,ev_empty=0,ev_cfeb[3]={0,0,0},ev_alct[3]={0,0,0},ev_tmb[3]={0,0,0},ev_one_csc=0;
  int ev_mult_csc=0,ev_mult_full_csc=0, trgwc_ok=0;
  unsigned int bytes,acrc_r,tcrc_r,crc22,crc_r,crc15;
  unsigned int l1num=0,trg_wc,trg_mark,n_cfeb_samples=16;
  unsigned int idmb=0;
  unsigned int ialct[3]={0,0,0};
  unsigned int itmb[3]={0,0,0};
  unsigned int icfeb[3]={0,0,0};
  unsigned int ndmb=0;
  unsigned int nalct=0;
  unsigned int ntmb=0;
  unsigned int ncfeb=0;
  unsigned int dmb_cfeb=0,prev_dmb=0,ndmb_w_cfeb=0;
  unsigned int evt_stat=0,evt_wc=0,istart=0,iend=99999999,trg_rd,run_stat=0;
  unsigned int ierr_ddu=0,ierr_ev_ddu=0,ierr_dmb=0;
  unsigned int ierr_ev_dmb=0,ierr_tmb=0,ierr_ev_tmb=0;
  unsigned int ierr_alct=0,ierr_ev_alct=0,ierr_cfeb=0,ierr_ev_cfeb=0,flag=0;
  unsigned int ierr_rd=0,ierr_ev_rd=0,evt_error=0,ev_err=0;
  unsigned int ifeb=0, cfebmov=0, ihit[6], ioverlap, n_cfeb_ts=1;
  int idmb_err,icfeb_err,ialct_err,itmb_err,iread_err,iwrite=1,iprnt=0;
  int bxn,dmb_bxn,tmb_bxn,alct_bxn,tmb_lctbxn, minus; //tmb_pretrigger_bxn
  int alct_bxn_l0, alct_bxn_l1, temp_alct, minusa1, minusa2, vartemp, em_sp;
  int minusa3, minusa4;
  unsigned int dmb_l1a,tmb_l1a,tmb_nfeb,tmb_nts,tmb_active,tmb_dump,tmb_mode;
  unsigned int alct_l1a,alct_nts,alct_active,alct_dump,alct_mode, incr = 0;
  unsigned int ibx_err=0,NoDMBhdr=0,dmbdav,cscerr[16],dmb_id;
  int ddu_in[16];
  float good_eff, exc_eff;
  int strip_flag_0, strip_flag_1, clct0_kdistrip, clct1_kdistrip, clct0_kcfeb;
  int clct0_kdistrip1=0, clct0_kdistrip2=0, clct1_kdistrip1=0;
  int clct1_kdistrip2=0, clct0_vpf=0, clct1_vpf=0;
  int clct0_kcfeb1=0, clct0_kcfeb2=0, clct1_kcfeb1=0, clct1_kcfeb2=0;
  int clct1_kcfeb, temp_clct, clct0_khalfstrip, clct1_khalfstrip, tbin;
  int clct0_khalfstrip1=0, clct0_khalfstrip2=0, clct1_khalfstrip1=0;
  int clct1_khalfstrip2=0;
  int timesample, counter, igot=0, notStart=1, BadWord=0;
  float error_marg;
  unsigned short frame = 0, cfeb_no = 0, happen = 0, chamber_flag = 0;
  bool tmb_trail_flag = false;
  int kdmb[2];
  int kalct[2];
  int ktmb[2];
  int kcfeb[2];

  int orbit=924,iGAP,GAPsize;  //Note: lowest 7-bits of 3564=0xDEC is 0x6C=108
                               //Note: lowest 7-bits of 924=0x39C is 0x1C=28
  int jGAP, offsign=1, p_sign=1, n_sign=1;
  int p_dmbdduoffset=4,n_dmbdduoffset=-124;
  int p_dmbtmboffset=1,n_dmbtmboffset=-1;   // -1 <= range <= 1
  int p_dmbalctoffset=4,n_dmbalctoffset=3;
  int p_ddutmboffset=-1,n_ddutmboffset=-1;  // was -5 <= range <= -3, but always == -2 now (or -1)
  int p_ddualctoffset=128,n_ddualctoffset=127; //why so big? cable? BELOW...
  
  FILE *fp,*fpout;
  char Filename[90];
  char hist_File_name[90];
  char RunName[80];
  char *RName;

  char chmbstr[][11]={"Both","CSC #1","CSC #2"};
  //Arrays to be used in the tree
  
  Int_t tmb1_lctbxni = 0;
  Int_t tmb2_lctbxni = 0;
  Int_t tmb1_bxni = 0;
  Int_t tmb2_bxni = 0;
  Int_t bxnnoi = 0;
  Int_t alct1_bxi = 0;
  Int_t alct2_bxi = 0;
  Int_t dmb1_presa;
  Int_t dmb2_presa;
  Int_t alct1_presa;
  Int_t alct2_presa;
  Int_t tmb1_presa;
  Int_t tmb2_presa;
  Int_t cfeb1_presa;
  Int_t cfeb2_presa;
  Float_t eff;
  Int_t good_ev=0;

  //Array

  unsigned short ar[16][96];
  unsigned short tempar[16][100];

 // ALCT BX offset:
  p_ddualctoffset=124; n_ddualctoffset=-3440;  //why so big?  cable ~75ns=3BX
  p_ddualctoffset=126; n_ddualctoffset=125;  //setting at UCLA or pre-SPS25ns?
  p_ddualctoffset=125; n_ddualctoffset=124;  //TB: set range to 124-125 for SPS25ns?

  GAPsize=(orbit&0x7F);

// iGAP=(bxn+p_dmbdduoffset-orbit>0);
// GAPfix: p_dmbdduoffset=p_dmbdduoffset-GAPsize;
//        n_dmbdduoffset=n_dmbdduoffset+GAPsize;

  crc15=0;
  crc_r=0;
  crc22=0;
  acrc_r=0;
  tcrc_r=0;

/*  printf(" %d arguments detected.\n",argc-1);
  for(i=1;i<argc;i++){
    printf("  %s%s",argv[i],(i<argc-1) ? " " : "\n");
  }
*/

  if(argc > 1){
    //    RunName=(char *)argv[1];
    
    RName=argv[1];
    
// JRG, doesn't work:    if(argc>2)iwrite=((int) argv[2])&0x000f;
    if(argc>2)iwrite=atoi(argv[2])&0x000f;

    if(argc>3)iprnt=atoi(argv[3])&0x000f;
    printf("   Run Name=%s, iwrite = %d, iprnt=%d\n",argv[1],iwrite,iprnt);
  }
  else{
    printf("Input Run Name: ");
    scanf("%s",RunName);
    RName = RunName;
  }
  //  sprintf (Filename,"run%s.dat",RunName);

  sprintf (Filename,"run%s.dat",RName);

  //  printf("   opening file %s \n",Filename);

  fp = fopen (Filename,"r");

  //fp=fopen("file.dat","r");

  // V.A. begin --> copy of the ROOT example
  sprintf(hist_File_name,"run%shist.root",RName);

  int countpat=0;

  // Create a new ROOT binary machine independent file.
  // Note that this file may contain any kind of ROOT objects, histograms,
  // pictures, graphics objects, detector geometries, tracks, events, etc..
  // This file is now becoming the current directory.

  TFile *hfile=0;
  if(iwrite > 0) hfile = new TFile(hist_File_name,"RECREATE","Demo ROOT file with histograms");

  /* Create some histograms, a profile histogram and an ntuple.  To use, in 
     ROOT, type .x tree.cpp.  A list will appear with all of the different          objects that appear in the root file.  To draw a histogram, type               {histogram name}.Draw(). 
  */

  TH1F *hpx   = new TH1F("hpx","This is the px distribution",1848,-0.5,923.5);

  TH1F *dmb_cfeb_1 = new TH1F("dmb_cfeb_1","Number of CFEBs for CSC1",7,-1,6);
  TH1F *dmb_cfeb_2 = new TH1F("dmb_cfeb_2","Number of CFEBs for CSC2",7,-1,6);
  TH2F *hpxpy = new TH2F("hpxpy","py vs px",40,-4,4,40,-4,4);

  TH2F *adchist_1 = new TH2F("adchist_1","timesample vs. normalized ADC values for CSC1",48,-0.5,16.5,100,0,1);
  TH2F *adchist_2 = new TH2F("adchist_2","timesample vs. normalized ADC values for CSC2",48,-0.5,16.5,100,0,1);

  TProfile *hprof = new TProfile("hprof","Profile of pz versus px",100,-4,4,0,20);
 
  TNtuple *ntuple = new TNtuple("ntuple","Demo ntuple","px:py:pz:random:i");
  
  /* The following creates a tree.  To use the tree, in ROOT, after running the     C++ program, type {tree name}->StartViewer().  A window will appear with       leaves representing each branch.  To plot a histogram, double click on the     leaf you want.
     From the root prompt, you can also type {tree name}.Draw("{branch name}").     Note:  For some of these branches, you will have to cut off the excess 
     meaningless values.  In order to do this, you will have to apply a cut.
     The command, in Root, is TCut cutname = "branchname (>, <, =, etc.) value"
     Then, type treename.Draw("branchname",cutname)
     You can even specify several cuts and use them at the same time by typing 
     treename.Draw("branchname",firstcut (&&, ||, +, etc.) secondcut)
  */

  TTree *t1 = new TTree ("t1", "Test Tree");

  TBranch *alct1_bx = t1->Branch("bxn_csc1_alct", &alct1_bxi, "alct1_bx/I");

  TBranch *alct2_bx = t1->Branch("bxn_csc2_alct", &alct2_bxi, "alct2_bx/I");

  TBranch *bxnno = t1->Branch("ddu_bxn", &bxnnoi, "bxnno/I");
  
  TBranch *tmb1_lctbxn = t1->Branch("bxn_lct_csc1_tmb", &tmb1_lctbxni, "tmb1_lctbxn/I");
 
  TBranch *tmb2_lctbxn = t1->Branch("bxn_lct_csc2_tmb", &tmb2_lctbxni, "tmb2_lctbxn/I");
 
  TBranch *tmb_l1a1 = t1->Branch("bxn_csc1_tmb", &tmb1_bxni, "tmb_l1a1/I");

  TBranch *tmb_l1a2 = t1->Branch("bxn_csc2_tmb", &tmb2_bxni, "tmb_l1a2/I");

  TBranch *alct_bxn_l0_1 = t1->Branch("alct_bxn_l0_1", &minusa1, "alct_bxn_l0_1/I");
  TBranch *alct_bxn_l0_2 = t1->Branch("alct_bxn_l0_2", &minusa2, "alct_bxn_l0_2/I");
  TBranch *alct_bxn_l1_1 = t1->Branch("alct_bxn_l1_1", &minusa3, "alct_bxn_l1_1/I");
  TBranch *alct_bxn_l1_2 = t1->Branch("alct_bxn_l1_2", &minusa4, "alct_bxn_l1_2/I");

  TBranch *dmb1_pres = t1->Branch("dmb_present_csc1", &dmb1_presa, "dmb1_pres/I");

  TBranch *dmb2_pres = t1->Branch("dmb_present_csc2", &dmb2_presa, "dmb2_pres/I");

  TBranch *alct1_pres = t1->Branch("alct_present_csc1", &alct1_presa, "alct1_pres/I");

  TBranch *alct2_pres = t1->Branch("alct_present_csc2", &alct2_presa, "alct2_pres/I");

  TBranch *tmb1_pres = t1->Branch("tmb_present_csc1", &tmb1_presa, "tmb1_pres/I");

  TBranch *tmb2_pres = t1->Branch("tmb_present_csc2", &tmb2_presa, "tmb2_pres/I");

  TBranch *cfeb1_pres = t1->Branch("num_cfeb_present_csc1", &cfeb1_presa, "cfeb1_pres/I");

  TBranch *cfeb2_pres = t1->Branch("num_cfeb_present_csc2", &cfeb2_presa, "cfeb2_pres/I");

  TBranch *clct0_kdistrip_1 = t1->Branch("clct0_kdistrip_1", &clct0_kdistrip1, "clct0_kdistrip_1/I");
  TBranch *clct0_kdistrip_2 = t1->Branch("clct0_kdistrip_2", &clct0_kdistrip2, "clct0_kdistrip_2/I");
  TBranch *clct1_kdistrip_1 = t1->Branch("clct1_kdistrip_1", &clct1_kdistrip1, "clct1_kdistrip_1/I");
  TBranch *clct1_kdistrip_2 = t1->Branch("clct1_kdistrip_2", &clct1_kdistrip2, "clct1_kdistrip_1/I");
  TBranch *clct0_khalfstrip_1 = t1->Branch("clct0_khalfstrip_1", &clct0_khalfstrip1, "clct0_khalfstrip_1/I");
  TBranch *clct0_khalfstrip_2 = t1->Branch("clct0_khalfstrip_2", &clct0_khalfstrip2, "clct0_khalfstrip_2/I");
  TBranch *clct1_khalfstrip_1 = t1->Branch("clct1_khalfstrip_1", &clct1_khalfstrip1, "clct1_khalfstrip_1/I");
  TBranch *clct1_khalfstrip_2 = t1->Branch("clct1_khalfstrip_2", &clct1_khalfstrip2, "clct1_khalfstrip_2/I");
  TBranch *clct0_kcfeb_1 = t1->Branch("clct0_kcfeb_1", &clct0_kcfeb1, "clct0_kcfeb_1/I");
  TBranch *clct0_kcfeb_2 = t1->Branch("clct0_kcfeb_2", &clct0_kcfeb2, "clct0_kcfeb_2/I");
  TBranch *clct1_kcfeb_1 = t1->Branch("clct1_kcfeb_1", &clct1_kcfeb1, "clct1_kcfeb_1/I");
  TBranch *clct1_kcfeb_2 = t1->Branch("clct1_kcfeb_2", &clct1_kcfeb2, "clct1_kcfeb_2/I");

  //  TBranch *efficiency = t1->Branch("Efficiency", &eff, "efficiency/F");
  //  TBranch *error_margin = t1->Branch("Error_margin", &error_marg, "error_margin/F");

  // Fill histograms randomly
  Float_t px, py, pz;
  for (Int_t i = 0; i < 10000; i++) {
     gRandom->Rannor(px,py); //px and py will be two gaussian random numbers
     pz = px * px + py * py;
     Float_t random = gRandom->Rndm(1);
     //hpx->Fill(px);
     hpxpy->Fill(px,py);
     hprof->Fill(px,pz);
     ntuple->Fill(px,py,pz,random,i);
  }

  printf("\n BEGIN Processing Run file  run%s.dat  \n",RName);
  ierr = 1;
  timesample = 0;
  kalct[0] = 0;
  kalct[1] = 0;
  kdmb[0] = 0;
  kdmb[1] = 0;
  ktmb[0] = 0;
  ktmb[1] = 0;
  kcfeb[0] = 0;
  kcfeb[1] = 0;
  clct0_vpf=0;
  clct1_vpf=0;

  while(ierr > 0){

    while( (ierr > 0) && (igot==0) ){
      ierr = fscanf(fp,"%d   %04x %04x %04x %04x\n",&num,&buf[3],&buf[2],&buf[1],&buf[0]);
      if(ierr <= 0)printf(" ierr=%d, END OF FILE\n",ierr);

// check that start of data is DDU Event Header, if not then SKIP to next word
      if (notStart){
	if( (num == 1) && ((buf[3]&0xf000) == 0x5000) ){
	  notStart=0;
	  igot=1;
	  if (BadWord>0)printf("  * DDU readout: there were %d extra words before the first event.\n",BadWord);
// for this, increment ierr_rd at end in case there are no good events
	}
	else{
	  BadWord++;
	}
      }
      else igot=1;
    }

    tempar[timesample][frame++] = buf[0];
    tempar[timesample][frame++] = buf[1];
    tempar[timesample][frame++] = buf[2];
    tempar[timesample][frame++] = buf[3];
 
    //    printf("Read in word %d:  %04x %04x %04x %04x\n",num,buf[3],buf[2],buf[1],buf[0]);

//    check for DDU HDR3:  DMB_CNT, DMB_DAV
   
     if ((num == 3)&&(num == istart + 2)){
      istart = 0;
      ndmb = (buf[0]&0x000f);
      dmbdav = buf[1]&0x00007FFF;
      if (ndmb > 15) ndmb=16;
      j = 0x00000001;
      k = -1;
      //      for (i = 0; i < ndmb; i++){
      for (i = 0; i < 15; i++){
	cscerr[i] = 0;
	ddu_in[i] = 0;
	while (ddu_in[i] == 0){
	  k++;
	  ddu_in[i] = (dmbdav&(j<<k));
	  if (k >= 15) ddu_in[i] = -1;
	}
	ddu_in[i] = k;
      }
     }
// Begin new event on num==1:
     
     if (num == 1){
      iend = 99999999;

//    check for DDU Event Header

      if ((buf[3]&0xf000) != 0x5000){
	printf("  *** New event, DDU readout:  Missing DDU Event Header!\n");
	if (iread_err == 0){
	  ierr_rd++;
	  ierr_ev_rd++;
	}
	iread_err = 1;
      }
    }

//    check for 1st DMB Header2 (must occur immediately after DDU header!)
   
     if (num == 4 && ((buf[3]&0xf000) != 0xa000) && ((buf[2]&0xf000) != 0xa000)&& ((buf[1]&0xf000) != 0xa000) && ((buf[0]&0xf000) != 0xa000)){
    
      NoDMBhdr = 1;
      ev_empty++;
      ev_err = 1;
//    printf("  *** Missing 1st DMB Header2!\n");
    }





//    check for DDU special bit
    if ((buf[3]&0x8000) > 0){



// ALCT trail found:
      if ((buf[2]&0xffff) == 0xde0d){
	trgwc_ok=0;
	frame = 0;
	timesample = 0;
	countpat++;
	if (idmb == 1){
	  kalct[0] = 1;
	  alct1_bxi = alct_bxn;
	}
	if (idmb == 2) alct2_bxi = alct_bxn;
	if (idmb == 1) minusa1 = alct_bxn-alct_bxn_l0; 
	if (idmb == 2){
	  minusa2 = alct_bxn - alct_bxn_l0;
	    kalct[1] = 1;
	}
	if (idmb == 1) minusa3 = alct_bxn - alct_bxn_l1;
	if (idmb == 2) minusa4 = alct_bxn - alct_bxn_l1;

//    Compare ALCT CRC results:
	ialct[0]++;
        if((idmb>0)&&(idmb<3))ialct[idmb]++;
	if (ialct[0] > nalct) printf("  *** too many ALCTs: read %d,  expected %d\n",ialct[0],nalct);
	acrc_r = (((buf[1]&0x07ff) << 11)|(buf[0]&0x07ff));
	trg_wc = (buf[3]&0x03ff);
	trg_rd = ((num-trg_mark) * 4);
	if (trg_wc != trg_rd){
	  printf("  *** ALCT wordcount error: read %d (0x%x), expected %d (0x%x)\n",trg_rd,trg_rd,trg_wc,trg_wc);
	  ialct_err++;
	}
	else {
	  trgwc_ok=1;
	}

	if (crc22 != acrc_r){
	  printf("  *** ALCT crc calculated=%06x,   received=%06x:  ",crc22&0x03fffff,acrc_r&0x03fffff);
	  printf("ALCT CRC failed! \n");
	  ialct_err++;
	}
	else if(iprnt>0){
	  if(trgwc_ok>0)printf("  ALCT WC=%d ok,",trg_wc);
	  printf("  ALCT crc ok!  calculated=%06x, received=%06x\n",crc22&0x03fffff,acrc_r&0x03fffff);
	}

	acrc_r = 0;
// dmb_l1a,dmb_bxn,alct_bxn,alct_l1a,alct_nts,alct_active,alct_dump,
//   alct_mode: 0=NoDump/1=FullDump/2=LocalDump
	if ((l1num&0x0000000f) != alct_l1a){
	  printf("  *** L1A mismatch DDU-ALCT:  DDU=0x%1x, ALCT=0x%1x\n",l1num&0x000f,alct_l1a);
	  ialct_err++;
	}
	if (ialct_err > 0){
	  ierr_alct++;
	  ierr_ev_alct++;
	  ialct_err = -999999;
	}

// Compare DMB BXN with ALCT BXN (7-bits)....NO!
// Compare DDU BXN with ALCT BXN (12-bits):
	trg_bx_offset = dmb_bxn - (alct_bxn&0x0000007f);
	trg_bx_offset = bxn - alct_bxn;
	//if (1){
       	if ((trg_bx_offset != n_ddualctoffset) && (trg_bx_offset != p_ddualctoffset) && ((orbit + trg_bx_offset) != n_ddualctoffset) && ((orbit + trg_bx_offset) != p_ddualctoffset)){
	  //printf("   ** BXN mismatch DMB-ALCT:  DMB=0x%02x, ALCT=0x%03x (dec%d), offset=%d\n",dmb_bxn,alct_bxn&0x0fff,alct_bxn&0x0fff,trg_bx_offset);
	  printf("   ** BXN mismatch DDU-ALCT:  DDU=0x%03x, ALCT=0x%03x (dec%d), offset=%d\n",bxn&0x0fff,alct_bxn&0x0fff,alct_bxn&0x0fff,trg_bx_offset);
	  printf("    * ALCT:  mode=%1d,  nTimeBins=%d,  active=0x%x,  dumped=0x%x\n",alct_mode,alct_nts,alct_active,alct_dump);
	}

      }

// TMB trail found:
// Filling histograms and tree

      else if ((buf[2]&0xffff) == 0xde0f){
	trgwc_ok=0;
	//	tmb_trail_flag = true;
	frame = 0;
	timesample = 0;
	if (idmb == 1) ktmb[0] = 1;
	if (idmb == 2) ktmb[1] = 1;
	if (idmb == 1) tmb1_lctbxni = tmb_lctbxn;
	if (idmb == 2) tmb2_lctbxni = tmb_lctbxn;
	if (idmb == 1) tmb1_bxni = tmb_bxn;
	if (idmb == 2) tmb2_bxni = tmb_bxn;
	if (idmb == 1)
	  {
	    if (strip_flag_0 == 0) clct0_kdistrip1 = clct0_kdistrip;
	    else clct0_khalfstrip1 = clct0_khalfstrip;
	    if (strip_flag_1 == 0) clct1_kdistrip1 = clct1_kdistrip;
	    else clct1_khalfstrip1 = clct1_khalfstrip;

	    if((clct0_kcfeb>=0)&&(clct0_kcfeb<6))clct0_kcfeb1 = clct0_kcfeb;
	    if((clct1_kcfeb>=0)&&(clct1_kcfeb<6))clct1_kcfeb1 = clct1_kcfeb;
	    if(iprnt>0)printf(" ***** CLCT0 Check:  clct0_kcfeb=%d\n",clct0_kcfeb);
	    if(iprnt>0)printf(" ***** CLCT1 Check:  clct1_kcfeb=%d\n",clct1_kcfeb);
	    clct0_kcfeb = 0;
	    clct1_kcfeb = 0;
	    clct0_vpf=0;
	    clct1_vpf=0;
	  }
	if (idmb == 2)
	  {
	    if (strip_flag_0 == 0) clct0_kdistrip2 = clct0_kdistrip;
	    else clct0_khalfstrip2 = clct0_khalfstrip;
	    if (strip_flag_1 == 0) clct1_kdistrip2 = clct1_kdistrip;
	    else clct1_khalfstrip2 = clct1_khalfstrip;

	    if((clct0_kcfeb>=0)&&(clct0_kcfeb<6))clct0_kcfeb2 = clct0_kcfeb;
	    if((clct1_kcfeb>=0)&&(clct1_kcfeb<6))clct1_kcfeb2 = clct1_kcfeb;
	    if(iprnt>0)printf(" ***** CLCT0 Check:  clct0_kcfeb=%d\n",clct0_kcfeb);
	    if(iprnt>0)printf(" ***** CLCT1 Check:  clct1_kcfeb=%d\n",clct1_kcfeb);
	    clct0_kcfeb = 0;
	    clct1_kcfeb = 0;
	    clct0_vpf=0;
	    clct1_vpf=0;
	  }

//    Compare TMB CRC results:
	itmb[0]++;
        if((idmb>0)&&(idmb<3))itmb[idmb]++;
	//	ktmb[t] = 1;
	if (itmb[0] > ntmb) printf("  *** too many TMBs: read %d,  expected %d\n",itmb[0],ntmb);
	tcrc_r = (((buf[1]&0x07ff) << 11)|(buf[0]&0x07ff));
	trg_wc = (buf[3]&0x07ff);
	trg_rd = ((num - trg_mark) * 4);
	if (trg_wc != trg_rd){
	  printf("  *** TMB wordcount error: read %d (0x%x), expected %d (0x%x)\n",trg_rd,trg_rd,trg_wc,trg_wc);
	  itmb_err++;
	}
	else {
	  trgwc_ok=1;
	}

	if (crc22 != tcrc_r){
	  printf("  *** TMB crc calculated=%06x,   received=%06x:  ",crc22&0x03fffff,tcrc_r&0x03fffff);
	  printf("TMB CRC failed! \n");
	  itmb_err++;
	}
	else if(iprnt>0){
	  if(trgwc_ok>0)printf("  TMB WC=%d ok,",trg_wc);
	  printf("  TMB crc ok!  calculated=%06x, received=%06x\n",crc22&0x03fffff,tcrc_r&0x03fffff);
	}

	tcrc_r = 0;
// dmb_l1a,dmb_bxn,tmb_bxn,tmb_l1a,tmb_nfeb,tmb_nts,tmb_active,tmb_dump,
//   tmb_mode:
//     0=NoDump/1=FullDump/2=LocalDump/3=NoDumpShortHdr/4=NoDumpNoHdr

	if ((l1num&0x0000000f) != tmb_l1a)
	  {
	  printf("  *** L1A mismatch DDU-TMB:  DDU=0x%1x, TMB=0x%1x\n",l1num&0x000f,tmb_l1a);
	  itmb_err++;
	  }
	if (itmb_err > 0)
	  {
	  ierr_tmb++;
	  ierr_ev_tmb++;
	  itmb_err = -999999;
	  }

// Compare DMB BXN with TMB BXN (one 7-bits, one 12-bits)....NO!
/*	trg_bx_offset=dmb_bxn-(tmb_bxn&0x0000007f);
	if((trg_bx_offset<n_dmbtmboffset) || (trg_bx_offset>p_dmbtmboffset))
	{
	  printf("   ** BXN mismatch DMB-TMB:  DMB=0x%02x, TMB=0x%03x (dec%d), offset=%d\n",dmb_bxn,tmb_bxn&0x0fff,tmb_bxn&0x0fff,trg_bx_offset);
	  ibx_err++;
	}
*/

// Compare DDU BXN with TMB BXN (12-bits):

	trg_bx_offset = bxn-tmb_bxn;

// JRG, 8/5/03:  Need to take SIGN of set_offset times orbit below:  -orbit
//	if (1){
//	if((trg_bx_offset<n_ddutmboffset) || (trg_bx_offset>p_ddutmboffset)){

       	if ((trg_bx_offset != n_ddutmboffset) && (trg_bx_offset != p_ddutmboffset) && ((trg_bx_offset - orbit) != n_ddutmboffset) && ((trg_bx_offset - orbit) != p_ddutmboffset))
	  {
	  printf("   ** BXN mismatch DDU-TMB:  DDU=0x%03x, TMB=0x%03x, offset=%d\n",bxn,tmb_bxn,trg_bx_offset);
	  ibx_err++;
	  }

	if (ibx_err > 0) printf("    * TMB:  mode=%1d,  nTimeBins=%d,  nfeb=%d,  active=0x%x,  dumped=0x%x\n",tmb_mode,tmb_nts,tmb_nfeb,tmb_active,tmb_dump);
	}

// DMB header found:

      else if (((buf[3]&0xf000) == 0xa000) && ((buf[2]&0xf000) == 0xa000) && ((buf[1]&0xf000) == 0xa000) && ((buf[0]&0xf000) == 0xa000)) {
//    check for DMB header
//
// for MPC: if next word has (buf[0]&0x0fff)=0x0b0c then it is TMB data!
//    TMB data continues through the "marker word" with 4-consecutive hex "Dxxx" codes,
//    the 3rd of which is "DEOF" and the 4th is TMB wordcount (lowest 11-bits)
//
	idmb++;
	frame = 0;
	timesample = 0;
	dmb_l1a = (buf[3]&0x000000ff);
	dmb_bxn = (buf[2]&0x0000007f);
	dmb_id = (buf[1]&0x0000000f);
	printf("   Begin CSC #%d of %d (DDU Input #%d, DMB ID=0x%X):  ",idmb,ndmb,ddu_in[idmb-1],dmb_id);
	if (idmb == 1) chamber_flag = 0;
	if (idmb == 2) chamber_flag = 1;
	if ((buf[0]&0x0800) > 0) ntmb++;
	if ((buf[0]&0x0400) > 0) nalct++;
	if (ddu_in[idmb-1] == 10){
	  kdmb[0] = 1;
	  frame = 0;
	  timesample = 0;
	}
	if (ddu_in[idmb-1] == 12){
	  kdmb[1] = 1;
	  frame = 0;
	  timesample = 0;
	}

// Count CFEB DAVs on the DMB:
	dmb_cfeb = 0;
	cfebmov = 0;
	for (i = 0; i < 5; i++){
	  ihit[i] = 0;
	  cfebmov = ((buf [2] & (0x0080 << i)) > 0);
	  if (cfebmov) printf(" CFEB#%1d: Multiple Overlap!\n",i+1);
	  if( (buf[0]&(0x0001<<i)) >0){
	    dmb_cfeb++;
	    ihit[i]++;
	  }
//	  printf("i=%d, ncfeb=%d\n",i,dmb_cfeb);
	}
	if (idmb == 1)
	  {
	    if (dmb_cfeb != 0) dmb_cfeb_1->Fill(dmb_cfeb);
	    dmb1_presa = kdmb[0];
	  }
	if (idmb == 2)
	  {
	    if (dmb_cfeb != 0) dmb_cfeb_2->Fill(dmb_cfeb);
	    dmb2_presa = kdmb[1];
	  }
	printf("nCFEB=%d, nALCT=%d, nTMB=%d\n",dmb_cfeb,(buf[0]&0x0400)>>10,(buf[0]&0x0800)>>11);
	printf("      cfeb_hit[5:1]=%1d%1d%1d%1d%1d\n",ihit[4],ihit[3],ihit[2],ihit[1],ihit[0]);
	ncfeb = ncfeb + dmb_cfeb;

// Compare DDU and DMB L1A (8-bits):
	if ((l1num&0x000000ff) != dmb_l1a){
	  printf("  *** L1A mismatch DMB-DDU:  DMB=0x%x, DDU=0x%x\n",dmb_l1a,l1num);
	  if (idmb_err == 0){
	    ierr_dmb++;
	    ierr_ev_dmb++;
	  }
	  idmb_err = 1;
	}

// Compare DDU and DMB BXN (7-bits):
// JRG, 10June03:  Need to take SIGN of set_offset times orbit below:  -orbit
	bx_offset=dmb_bxn-(bxn&0x0000007f);
	offsign=1;
	if(bx_offset<0)offsign=-1;
	iGAP=(bxn+p_dmbdduoffset-orbit>=0);
	jGAP=(bxn+n_dmbdduoffset-orbit>=0);
	p_sign=1;
	n_sign=1;
	if(p_dmbdduoffset<0){
	  iGAP=(bxn+p_dmbdduoffset<0);
	  p_sign=-1;
	}
	if(n_dmbdduoffset<0){
	  jGAP=(bxn+n_dmbdduoffset<0);
	  n_sign=-1;
	}

	bx_offset = dmb_bxn - (bxn&0x0000007f);
	iGAP = (bxn + p_dmbdduoffset - orbit>=0);
 
	if ((bx_offset != (n_dmbdduoffset + GAPsize * iGAP)) && (bx_offset != (p_dmbdduoffset - GAPsize * iGAP))){
	  printf("  *** BXN mismatch DMB-DDU:  DMB=0x%02x, DDU=0x%03x (dec%d), offset=%d\n",dmb_bxn,bxn&0x0fff,bxn&0x0fff,bx_offset);
	}
      }


// Check for event end marker (actually "DDU Trail-2"):
      else if (((buf[3]&0xffff) == 0x8000) && ((buf[2]&0xffff) == 0xffff) && ((buf[1]&0xffff) == 0x8000) && ((buf[0]&0xffff) == 0x8000)) {
	iend = num;
        for(l=0;l<3;l++){
	  if (icfeb[l] > 0) ev_cfeb[l]++;
	  if (ialct[l] > 0) ev_alct[l]++;
	  if (itmb[l] > 0) ev_tmb[l]++;
        }
	if (idmb == 1) ev_one_csc++;
	if (idmb > 1) ev_mult_csc++;

	printf("  Event Summary (ev %d=0x%x): \n ",l1num,l1num);
	if (idmb != ndmb){
	  printf("  *** DMB count mismatch!  read %d, expected %d.\n",idmb,ndmb);
	  evt_error++;
	  ev_err = 1;
	}
	if (ialct[0] < nalct){
	  printf("  *** missing ALCT data!  Expected %d but only got %d. \n",nalct,ialct[0]);
	  evt_error++;
	  ev_err = 1;
	}
	if (itmb[0] < ntmb){
	  printf("  *** missing TMB data!  Expected %d but only got %d. \n",ntmb,itmb[0]);
	  evt_error++;
	  ev_err = 1;
	}
// JRG, need to generalize for 8 time samples:
	if ((icfeb[0]/n_cfeb_samples) != ncfeb){
	  printf("  *** Missing CFEB samples!  read %d samples, expected %d.\n",icfeb[0],ncfeb*n_cfeb_samples);
	  evt_error++;
	  ev_err = 1;
	}
	else if (((icfeb[0]/n_cfeb_samples) * n_cfeb_samples) < icfeb[0]){
	  printf("  *** Extra CFEB samples!  read %d samples, expected %d.\n",icfeb[0],ncfeb*n_cfeb_samples);
	  evt_error++;
	  ev_err = 1;
	}
	if ((ierr_tmb > 0)||(ierr_dmb > 0)||(ierr_alct > 0)||(ierr_ddu > 0)||(ierr_cfeb > 0)||(ierr_rd > 0))
	{
	  evt_error++;
	}
	if ((ierr_ev_tmb > 0)||(ierr_ev_dmb > 0)||(ierr_ev_alct > 0)||(ierr_ev_ddu > 0)||(ierr_ev_cfeb > 0)||(ierr_ev_rd > 0))
	  {
	  if (flag == 0)
	    {
	    ev_err = 1;
	    flag++;
	    }
	  }

	if ((idmb > 1) && (ialct[0] > 1) && (itmb[0] > 1) && (ndmb_w_cfeb > 1)){
// && (evt_error == 0)
	  ev_mult_full_csc++;
	  printf("Excellent Event!  %d fully occupied CSCs!\n",ndmb_w_cfeb);
	}
	else printf("\n");
	printf("  * %d CFEBs,  %d DMB, %d ALCT, %d TMB,  %d DDU words read (ev %d=0x%x)\n",icfeb[0]/n_cfeb_samples,idmb,ialct[0],itmb[0],num+2,l1num,l1num);

	if (ev_err != 1) good_ev++;
	ndmb = 0;
	ndmb_w_cfeb = 0;
	nalct = 0;
	ntmb = 0;
	ncfeb = 0;
	idmb = 0;
        for(l=0;l<3;l++){
	  ialct[l] = 0;
	  itmb[l] = 0;
	  icfeb[l] = 0;
	}
	tmb1_presa = ktmb[0];
	tmb2_presa = ktmb[1];
	cfeb1_presa = kcfeb[0];
	cfeb2_presa = kcfeb[1];
	alct1_presa = kalct[0];
	alct2_presa = kalct[1];
	t1->Fill();
      }


// Check for event end (actually "DDU Trail" because iend=num at Trail-2):
   
      else if (num == iend + 2){

//  check for event wordcount
	evt_wc = ((buf[3]&0x00ff) << 16)|(buf[2]&0xffff);
	if (num != evt_wc){
	  printf("  *** Event wordcount error!  DDU readout %d (0x%x), expected %d (0x%x)\n",num,num,evt_wc,evt_wc);
	  if (iread_err == 0){
	    ierr_rd++;
	    ierr_ev_rd++;
	  }
	  iread_err = 1;
	}
	iend = 88888888;
//	istart=0;
// JRG, zero some stuff at the end-of-event (after tree Fill):
	clct0_kdistrip1 = 0;
	clct1_kdistrip1 = 0;
	clct0_kdistrip2 = 0;
	clct1_kdistrip2 = 0;
	clct0_khalfstrip1 = 0;
	clct1_khalfstrip1 = 0;
	clct0_khalfstrip2 = 0;
	clct1_khalfstrip2 = 0;
	clct0_kcfeb1 = 0;
	clct1_kcfeb1 = 0;
	clct0_kcfeb2 = 0;
	clct1_kcfeb2 = 0;
	clct0_vpf=0;
	clct1_vpf=0;
      }

// Check for ethernet "filler" junk passed by gigabit driver; should only
//  happen for empty events at end of small event packet with 48 bytes!

      else if (((buf[3]&0xffff) == 0xffff) && ((buf[2]&0xffff) == 0xffff) && ((buf[1]&0xffff) == 0xffff) && ((buf[0]&0xff00) == 0xff00)) {
	bytes = (buf[0]&0x000000ff);

	if (iend == 99999990){

//    Event end should be previous word (num==6) so now check num==7

	  if (num != 7) printf("   ** Empty event detected with filler, but wordcount!=6: %d\n",num-1);

	  if (bytes != 48) printf("   ** Empty event detected with filler, but data bytes!=48: %d\n",bytes);

	  iend = 99999994;
	}

//   Notify if filler junk happens in any other case: error!?

	else if (iend != 99999994){
	  if (NoDMBhdr > 0) printf("  *** Missing 1st DMB Header2!\n");
	  printf("  *** Ethernet Filler junk detected");
	  if (iend != 88888888) printf(" mid-event");
	  if (bytes == 255){
	    printf("!  \n");
	  }

	  else{
	    printf("!  %d data bytes present\n",bytes);
	  }

	  iend = 99999994;
	}
      }

// Notify if normal empty event was detected (regardless of "filler" at end):

      if ((iend == 88888888) && (NoDMBhdr > 0)){
	if (num == 6){
	  printf("    * Empty event detected. \n");
	}

	else{
	  printf("   ** Empty event detected, but wordcount!=6: %d\n",num);
	}

	iend = 99999990;
      }

// Notify if No DMB Header detected but event is Not Empty:

      if ((iend < 99999990) && (num > 6) && (NoDMBhdr > 0)){
	printf("  *** Missing 1st DMB Header2 in a big event!\n");
      }

      crc15 = 0;

// Zero CRC22 accumulator for any non-"B" DDU special word (after CRC check)

      if (((buf[3]&0xf000) == 0xb000) && ((buf[0]&0xf000) == 0xb000)){
	printf("  *** DMB Error Code detected! \n");
	if (icfeb_err == 0){
	  ierr_cfeb++;
      	  ierr_ev_cfeb++;
	}
	icfeb_err = 1;
      }

      else{
	crc22 = 0;
	trg_mark = num;
// JRG, zero some stuff at trg_mark points:
	clct0_kcfeb = 0;
	clct1_kcfeb = 0;
	clct0_vpf=0;
	clct1_vpf=0;
      }
    }



//  if no DDU special bit, check if first word of event, header info:

    else if ((num == 1) && (buf[3]&0xf000) == 0x5000){

      ev++;
      frame = 0;
      timesample = 0;
      happen = 0;
      tmb_trail_flag = false;
// JRG 28 June 03, move this to end-of-event:     if (ev_err != 1) good_ev++;
      istart = num;
      ntmb = 0;
      nalct = 0;
      for(l=0;l<3;l++){
        ialct[l] = 0;
        itmb[l] = 0;
        icfeb[l] = 0;
      }
      ncfeb = 0;
      ndmb = 0;
      ndmb_w_cfeb = 0;
      idmb = 0;
      ev_err = 0;
      ierr_ev_tmb = 0;
      ierr_ev_ddu = 0;
      ierr_ev_dmb = 0;
      ierr_ev_cfeb = 0;
      ierr_ev_alct = 0;
      ierr_ev_rd = 0;
      flag = 0;
      idmb_err = 0;
      icfeb_err = 0;
      ialct_err = 0;
      itmb_err = 0;
      iread_err = 0;
      evt_error = 0;
      clct0_kdistrip1 = 0;
      clct1_kdistrip1 = 0;
      clct0_kdistrip2 = 0;
      clct1_kdistrip2 = 0;
      clct0_khalfstrip1 = 0;
      clct1_khalfstrip1 = 0;
      clct0_khalfstrip2 = 0;
      clct1_khalfstrip2 = 0;
      clct0_kcfeb1 = 0;
      clct1_kcfeb1 = 0;
      clct0_kcfeb2 = 0;
      clct1_kcfeb2 = 0;
      clct0_vpf=0;
      clct1_vpf=0;
      ktmb[0] = 0;
      ktmb[1] = 0;
      kdmb[0] = 0;
      kdmb[1] = 0;
      kcfeb[0] = 0;
      kcfeb[1] = 0;
      kalct[0] = 0;
      kalct[1] = 0;
      NoDMBhdr = 0;
      l1num = ((buf[3]&0x00ff) << 16)|(buf[2]&0xffff);
      bxn = ((buf[1]&0x0000fff0) >> 4);
      bxnnoi = bxn;
      ibx_err = 0;
      printf("ev#%d:  LEVEL 1 Event Number = %d (0x%x)\n",ev,l1num,l1num);
      if (l1num != ev+l1offset){
	prev_offset = l1offset;
	l1offset = l1num-ev;
	printf("   ** L1A number jump:  %d event numbers skipped\n",l1offset-prev_offset);
      }
    }


// Check for CFEB timesample end, check CFEB CRC and error code?
//    else if (((buf[3]&0xffff) == 0x7fff) && ((buf[2]&0xf000) == 0x7000) && ((buf[1]&0xf000) == 0x7000)){
    else if(((buf[3]&0xffff)==0x7fff) && ((((num-trg_mark)/25)*25)==(num-trg_mark))){
      
      for (counter=0;counter<96;counter++){
	ar[timesample][counter] = tempar[timesample][counter];
      }
	if (((tempar[timesample][95])&0x2000) == 0) n_cfeb_samples = 8;
	else n_cfeb_samples = 16;
      if (((timesample%(n_cfeb_samples-1))==0)&&(timesample!=0)){
	ADChistos(ar,adchist_1,adchist_2,n_cfeb_samples,chamber_flag);
      }
      icfeb[0]++;
      if((idmb>0)&&(idmb<3))icfeb[idmb]++;
      if (((timesample%(n_cfeb_samples-1))==0)&&(timesample!=0)) timesample = 0;
      else timesample++;
      frame = 0;
      // n_cfeb_samples=8+(n_cfeb_ts*8);

      if ((timesample%n_cfeb_samples) == 0){
	if (idmb == 1) kcfeb[0]++;
	if (idmb == 2) kcfeb[1]++;
      }
      if (idmb != prev_dmb){
	ndmb_w_cfeb++;
	prev_dmb = idmb;
      }
      crc_r = (buf[0]&0x7fff);

      if (crc15 != crc_r){
	printf("  *** CFEB CRC failed!  calculated=%04x,   received=%04x:  ",crc15&0x07fff,crc_r&0x7fff);
	printf("cfeb timesample #%d\n",icfeb[0]);
	if (ioverlap) printf(" OVLP");
	if (((crc15&0x4000)==0)^(ioverlap)) printf(" FLP\n");
	else printf("\n");
	if (icfeb_err == 0){
	  ierr_cfeb++;
       	  ierr_ev_cfeb++;
	}
	icfeb_err = 1;
      }
      else if(iprnt>0){
	printf("  CFEB crc ok!  calculated=%04x, received=%04x:  ",crc15&0x07fff,crc_r&0x7fff);
	printf("cfeb timesample #%d",icfeb[0]);
	if (ioverlap)printf(" OVLP");
	if ( ((crc15&0x4000)==0)^(ioverlap) )printf(" FLP\n");
	else printf("\n");
      }

/*      else{
	printf("CFEB CRC passed!  cfeb timesample #%d\n",icfeb[0]);
      }
*/
      crc15 = 0;
    }

    else if (num == iend + 1){

//  check for event status

      evt_stat = ((buf[3]&0xffff) << 16)|(buf[2]&0xffff);

//JRG, 24may03:  Temp mask for DDU v8r15576 bug in Trig L1A comparison
//  REMOVE MASK LATER when ddu bug is fixed!

      //      if (((evt_stat&0x000D0000) > 0) && ((evt_stat&0xFFF2BFFF) == 0)) evt_stat= 0;

//1st try at MASK:      if ((evt_stat != 0) && ((evt_stat&0x000D0000) == 0)){
      // Toni mark

      run_stat=run_stat|evt_stat;

      //      if (((evt_stat&0x000E0000)>0) && ((evt_stat&0xFFF1BEFF)==0)) evt_stat=0;
      if (evt_stat != 0) {
	run_stat = run_stat|evt_stat;
	printf("   ** Non-zero DDU event status:   %04X/%04X\n",((evt_stat&0xffff0000)>>16),(evt_stat&0x0000ffff));
	ddu_status_decode(evt_stat);
	ierr_ddu++;
	ierr_ev_ddu++;
      }
    }


// for normal data, Accumulate CRC15 and Trigger CRC22:

    else {
      for (l = 0; l < 4; l++) crc15 = (buf[l]&0x1fff)^((buf[l]&0x1fff) << 1)^(((crc15&0x7ffc) >> 2)|((0x0003&crc15) << 13))^((crc15&0x7ffc) >> 1);
      for (l = 0; l < 4; l++) crc22 = (buf[l]&0xffff)^((buf[l]&0xffff) << 1)^(((crc22&0x3fffc0) >> 6)|((0x00003f&crc22) << 16))^((crc22&0x3fffc0) >> 5);

      ioverlap = (((buf[2]&0x4000)==0) && ((buf[1]&0x4000)==0));
// bit 13 at end of every Time Sample defines the # of timesamples per event
//  1=16 TS, 0=8 TS

      //      n_cfeb_ts = (((buf[2]&0x2000)==1) && ((buf[1]&0x2000)==1));

      if ((num > istart + 2) && ((num - trg_mark) == 1)){

//  get trigger event info available in 1st word of dump:
//                                       tmb_bxn,tmb_l1a,tmb_nfeb,tmb_nts;

	tmb_bxn = (buf[3]&0x00000fff);

	//   tmb_pretrigger_bxn=(buf[7]&0x00000fff);

	tmb_l1a = (buf[2]&0x0000000f);

/* old TMB:
	tmb_nfeb = (buf[1]&0x00001c00) >> 10;
	tmb_nts = (buf[1]&0x000003e0) >> 5;
*/

/* E-series TMB: */

	tmb_mode = (buf[1]&0x00007000) >> 12;
	tmb_dump = (buf[1]&0x00000fe0) >> 5;
	tmb_nts = (buf[1]&0x0000001f);

	alct_dump = (buf[3]&0x00003f80)>>7;
	alct_active = (buf[3]&0x0000007f);
	alct_bxn = (buf[2]&0x00000fff);

//	alct_lctbxn = (buf[2]&0x00000fff);

	alct_nts = (buf[1]&0x0000007c) >> 2;
	alct_mode = (buf[1]&0x00000003);
	alct_l1a = (buf[0]+1)&0x0000000f;  //Madorsky started with evt #0...
	alct_l1a = (buf[0]&0x0000000f);
      }
      else if ((num - trg_mark) == 2){

//  get trigger event info available in 2nd word of dump:
//                                       tmb_active,tmb_dump,tmb_mode;

/* old TMB:
	tmb_active=(buf[1]&0x0000001f);
	tmb_dump=(buf[0]&0x00000f80)>>7;
	tmb_mode=(buf[0]&0x00000003);
*/

/* E-series TMB: */

	tmb_nfeb = (buf[0]&0x000000e0) >> 5;
	tmb_active = (buf[2]&0x0000001f);
	tmb_lctbxn = (buf[3]&0x00000fff);

	temp_alct = (buf[0]&0x00007000) >> 12;
	alct_bxn_l0 = (buf[1]&0x00000003) << 3;
	alct_bxn_l0 = alct_bxn_l0 + temp_alct;

	/*	if (ev < 6){
	  //	  print("buffer[1]: %\n",buf[1]);
	  printf("alct_bxn_l0: %d\n",alct_bxn_l0);
	  printf("temp_alct: %d\n",temp_alct);
	  }*/

	temp_alct = (buf[2]&0x00007000) >> 12;
	alct_bxn_l1 = (buf[3]&0x00000003) << 3;
	alct_bxn_l1 = alct_bxn_l1 + temp_alct;

	/*	if (ev < 6){
	  printf ("alct_bxn_l1: %d\n", alct_bxn_l1);
	  printf ("temp_alct: %d\n", temp_alct);
	  }*/

      }

        else if ((num - trg_mark) == 3){
//  get CLCT info available in 3rd word of dump:
//      for LCT0 and LCT1, find key CFEB & key STRIP if Valid Pattern Flag
	  clct0_vpf = (buf[0]&0x00000001);
	  if (clct0_vpf>0){
	    temp_clct = (buf[0]&0x00004000) >> 14;
	    clct0_kcfeb = (buf[2]&0x00000003) << 1;
	    clct0_kcfeb = clct0_kcfeb + temp_clct + 1;
	    if((clct0_kcfeb>5)&&(iprnt>0))printf(" **** CLCT0 Too BIG:  clct0_kcfeb=%d\n",clct0_kcfeb);

	    strip_flag_0 = (buf[0]&0x00000080) >> 7;
	    if (strip_flag_0 == 0){
	      clct0_kdistrip = (buf[0]&0x00003e00) >> 9;
	      clct0_kdistrip = clct0_kdistrip/4;
	      clct0_kdistrip = (8 * (clct0_kcfeb-1)) + clct0_kdistrip;
	    }
	    else{
	      clct0_khalfstrip = (buf[0]&0x00003e00) >> 9;
	      clct0_khalfstrip = (32 * (clct0_kcfeb-1)) + clct0_khalfstrip;
	    }
	  }

	  clct1_vpf = (buf[1]&0x00000001);
	  if (clct1_vpf>0){
	    temp_clct = (buf[1]&0x00004000) >> 14;
	    clct1_kcfeb = (buf[2]&0x000000c0) >> 5;
	    clct1_kcfeb = clct1_kcfeb + temp_clct + 1;
	    if((clct1_kcfeb>5)&&(iprnt>0))printf(" **** CLCT1 Too BIG:  clct1_kcfeb=%d\n",clct1_kcfeb);

	    strip_flag_1 = (buf[1]&0x00000080) >> 7;
	    if (strip_flag_1 == 0){
	      clct1_kdistrip = (buf[1]&0x00003e00) >> 9;
	      clct1_kdistrip = clct1_kdistrip/4;
	      clct1_kdistrip = (8 * (clct1_kcfeb-1)) + clct1_kdistrip;
	    }
	    else{
	      clct1_khalfstrip = (buf[1]&0x00003e00) >> 9;
	      clct1_khalfstrip = (32 * (clct1_kcfeb-1)) + clct1_khalfstrip;
	    }
	  }

	}
    }

    igot=0;
  }


// Save all ROOT objects in this file
  eff = (1.0 * ev_mult_full_csc)/good_ev;
  error_marg = (sqrt(1.0 * ev_mult_full_csc))/good_ev;

  if (iwrite > 0) (*hfile).Write();

  // Close the file. Note that this is automatically done when you leave
  // the application.

  if (iwrite > 0) (*hfile).Close();

  // V.A. end

  fclose(fp);
  printf("\nProcessed %d events:  %d empty,  %d single-CSC,  %d multi-CSC\n",ev,ev_empty,ev_one_csc,ev_mult_csc);

  printf("    %d events w/CFEB, %d events w/ALCT, %d events w/TMB\n",ev_cfeb[0],ev_alct[0],ev_tmb[0]);
  printf("\nChamber   :    CFEB's    ALCT's     TMB's\n------------------------------------------------\n");
  for(l=0;l<3;l++){
    printf("%10s: %8d  %8d  %8d\n",chmbstr[l],ev_cfeb[l],ev_alct[l],ev_tmb[l]);
  }
  printf("\n");

  good_eff=0.;
  exc_eff=0.;
  if(ev>0)good_eff=1.0*good_ev/ev;
  if(ev>0)exc_eff=1.0*ev_mult_full_csc/ev;
//  if(good_ev>0)exc_eff=100.0*ev_mult_full_csc/good_ev;
  //  printf("    exc=%5.2f\%\    good=%5.2f\%\ \n",exc_eff,good_eff);

  if (ev_mult_full_csc <= 0) printf("  *** There were NO Excellent events and %d Good events\n",good_ev);
  if (ev_mult_full_csc > 0) printf("    There were %d Excellent events (%5.4f full data for multiple CSCs),\n               %d Good events (%5.4f)\n",ev_mult_full_csc,exc_eff,good_ev,good_eff);

  if ((run_stat != 0)||(ierr_cfeb != 0)||(ierr_alct != 0)||(ierr_tmb != 0)||(ierr_dmb != 0)||(ierr_rd != 0)){

//Toni mark:
    printf("    %d ev w/CFEB err, %d ev w/DMB err, %d ev w/ALCT err, %d ev w/TMB err\n",ierr_cfeb,ierr_dmb,ierr_alct,ierr_tmb);

// If there was memory buffer garbage at the start of run, increment error:
    if (BadWord>0)ierr_rd++;

    printf("  *** %d events had a DDU readout error\n",ierr_rd);

    if (run_stat != 0){
      printf("  *** %d events ended with a DDU detected error state\n",ierr_ddu);
      printf("  *** run summary of Non-zero DDU status:   %04X/%04X\n",((run_stat&0xffff0000)>>16),(run_stat&0x0000ffff));

      ddu_status_decode(run_stat);
      //      printf(" countpat %d \n",countpat); 
    }
  }
}
